/*
 * Command for accessing SPI flash.
 *
 * Copyright (C) 2008 Atmel Corporation
 * Licensed under the GPL-2 or later.
 */

#include <common.h>
#include <spi_flash.h>

#include <asm/io.h>

#ifndef CONFIG_SF_DEFAULT_SPEED
# define CONFIG_SF_DEFAULT_SPEED	10000000
#endif
#ifndef CONFIG_SF_DEFAULT_MODE
# define CONFIG_SF_DEFAULT_MODE		SPI_MODE_3
#endif

#define CONFIG_SF_READONLY

#define TagModel					1
#define TagRevision					2
#define TagManufacturer				3
#define TagSerialNumber				4
#define TagManufactureDate			5
#define TagNumEthPorts				6
#define TagNumSerialPorts			7
#define TagNumIOLinkPorts			8
#define TagMACAddr					9
#define TagDefaulIpAddr				10
#define TagDefaulNetmask			11
#define TagDefaulGatewayIp			12
#define TagTLANum					13
#define TagODVAVendorId				14
#define TagNumDioPorts				15
#define TagBTWWorkOrderNum			16

#define MAX_SYSINFO_ENV_NAME_LEN	20
#define NUM_SYSINFO_ENV				(sizeof(sysinfo_env)/sizeof(sysinfo_env_t))

typedef struct _sysinfo_env {
	int tag;
	char name[MAX_SYSINFO_ENV_NAME_LEN];
} sysinfo_env_t;

static sysinfo_env_t sysinfo_env[] = {
	{TagModel, "model"},
	{TagRevision, "revision"},
	{TagManufacturer, "manufacturer"},
	{TagSerialNumber, "serial#"},
	{TagManufactureDate, "manufacturedate"},
	{TagNumEthPorts, "numethports"},
	{TagNumSerialPorts, "numserialports"},
	{TagNumIOLinkPorts, "numiolinkports"},
	{TagMACAddr, "ethaddr"},
	{TagDefaulIpAddr, "defaultipaddr"},
	{TagDefaulNetmask, "defaultnetmask"},
	{TagDefaulGatewayIp, "defaultgatewayip"},
	{TagTLANum, "tlanum"},
	{TagODVAVendorId, "odvavendorid"},
	{TagNumDioPorts, "numdioports"},
	{TagBTWWorkOrderNum, "btwworkordernum"}
};

static struct spi_flash *flash;

static int do_spi_flash_probe(int argc, char *argv[])
{
	unsigned int bus = 0;
	unsigned int cs;
	unsigned int speed = CONFIG_SF_DEFAULT_SPEED;
	unsigned int mode = CONFIG_SF_DEFAULT_MODE;
	char *endp;
	struct spi_flash *new;

	if (argc < 2)
		goto usage;

	cs = simple_strtoul(argv[1], &endp, 0);
	if (*argv[1] == 0 || (*endp != 0 && *endp != ':'))
		goto usage;
	if (*endp == ':') {
		if (endp[1] == 0)
			goto usage;

		bus = cs;
		cs = simple_strtoul(endp + 1, &endp, 0);
		if (*endp != 0)
			goto usage;
	}

	if (argc >= 3) {
		speed = simple_strtoul(argv[2], &endp, 0);
		if (*argv[2] == 0 || *endp != 0)
			goto usage;
	}
	if (argc >= 4) {
		mode = simple_strtoul(argv[3], &endp, 16);
		if (*argv[3] == 0 || *endp != 0)
			goto usage;
	}

	new = spi_flash_probe(bus, cs, speed, mode);
	if (!new) {
          printf(" failed to initialize SPI flash at %u:%u %uHz mode %u\n", bus, cs, speed, mode);
          return 1;
	}

	if (flash)
		spi_flash_free(flash);
	flash = new;

	printf(" %u KiB %s at %u:%u is now current device\n",
			flash->size >> 10, flash->name, bus, cs);

	return 0;

usage:
	puts("Usage: sf probe [bus:]cs [hz] [mode]\n");
	return 1;
}

/**
* load_sysinfo_image - load legacy format System Information image
* @flash: pointer to current serial flash device
*
* load_sysinfo_image() verifies legacy image integrity and returns pointer to
* legacy image header if image verification was completed successfully.
*
* returns:
*     pointer to a legacy image header if valid image was found
*     otherwise return NULL
*/
static image_header_t *load_sysinfo_image(struct spi_flash *flash, int verbose)
{
	ulong offset = CONFIG_SPI_FLASH_SYSINFO_OFFSET;
	ulong addr = CONFIG_SYS_LOAD_ADDR;
	size_t cnt = image_get_header_size();
	image_header_t *hdr;
	int ret;

	// System Information is stored at different locations for
	// different serial flash chips. For MX25L512, the address
	// is 0x8000. For M25P40, the address is 0x10000
	if (!strcmp(flash->name, "M25P40")) {
		offset = CONFIG_SPI_FLASH_SYSINFO_OFFSET2;
	}
	printf("Loading SysInfo from serial flash at 0x%x\n", offset);

	/* Read Image Header into memory */
	ret = spi_flash_read(flash, offset, cnt, (void *)addr);
	if (ret) {
		printf(" error reading SPI Flash\n");
		return NULL;
	}

	/* Check Image Type */
	if (genimg_get_format((void *)addr) != IMAGE_FORMAT_LEGACY) {
		printf(" wrong image format\n");
		return NULL;
	}
	if (verbose)
		printf(" found System Information from Legacy Image at %x\n", CONFIG_SYS_LOAD_ADDR);

	/* Check Image Header */
	hdr = (image_header_t *)addr;
	if (verbose)
		image_print_contents(hdr);
	if (!image_check_hcrc(hdr)) {
		printf(" bad header checksum\n");
		return NULL;
	}
	if (image_get_type(hdr) != IH_TYPE_STANDALONE) {
		printf(" wrong image type\n");
		return NULL;
	}

	/* Read Image Data into memory */
	addr += image_get_header_size();
	offset += image_get_header_size();
	cnt = image_get_data_size(hdr);
	ret = spi_flash_read(flash, offset, cnt, (void *)addr);
	if (ret) {
		printf(" error reading SPI Flash\n");
		return NULL;
	}
	/* Check Image Data CRC Checksum */
	if (verbose)
		printf(" verifying checksum ... ");
	if (!image_check_dcrc(hdr)) {
		printf(" bad data CRC\n");
		return NULL;
	}
	if (verbose)
		printf(" OK\n");

	return hdr;
}

static char *get_name_by_tag(int tag)
{
	int i;
	for (i = 0; i < NUM_SYSINFO_ENV; ++i) {
		if (tag == sysinfo_env[i].tag)
			return sysinfo_env[i].name;
	}
	return NULL;
}

int is_sysinfo_env(char *name)
{
	int i;
	for (i = 0; i < NUM_SYSINFO_ENV; ++i) {
		if (strcmp(name, sysinfo_env[i].name) == 0)
			return 1;
	}
	return 0;
}

// If <name> is unset, set <name> to <value> return 1 if env changed,
// 0 if not changed.
static int setenv_default(char *name, char *value)
{
  char *curval = getenv(name);

  if (curval)
    return 0;

  printf(" %s<=%s\n",name,value);
  setenv(name, value);
  return 1;
}

int update_sysinfo_from_image(ulong image_data, uint32_t image_size, int verbose)
{
	uint8_t *data = (uint8_t *)image_data;
	uint8_t *data_end = data + image_size;
	uint16_t tag, length;
	char str[256];
	char *curval, *name;
	int changed = 0;

	while (data < data_end) {
		tag    = (data[0]<<8) | data[1];
		length = (data[2]<<8) | data[3];
		data += 4;
		name = get_name_by_tag(tag);
		if (!name) {
			printf(" unsupported tag %d(0x%x), length %d(0x%x)\n",
					tag, tag, length, length);
			data += length;
			continue;
		}
		if ((((tag == TagNumEthPorts) ||
			  (tag == TagNumSerialPorts) ||
			  (tag == TagNumIOLinkPorts) ||
			  (tag == TagNumDioPorts)) && (length != 1)) ||
			((tag == TagMACAddr) && (length != 6)) ||
			(((tag == TagDefaulIpAddr) ||
			  (tag == TagDefaulNetmask) ||
			  (tag == TagDefaulGatewayIp)) && (length != 4))) {
			printf(" invalid length %d(0x%x) of tag %d(0x%x)\n",
				   length, length, tag, tag);
			data += length;
			continue;
		}
		str[0] = '\0';
		switch (tag) {
			case TagNumEthPorts:
			case TagNumSerialPorts:
			case TagNumIOLinkPorts:
			case TagNumDioPorts:
				sprintf(str, "%d", data[0]);
				break;
			case TagMACAddr:
				sprintf(str, "%02x:%02x:%02x:%02x:%02x:%02x",
						data[0], data[1], data[2],
						data[3], data[4], data[5]);
				break;
			case TagDefaulIpAddr:
			case TagDefaulNetmask:
			case TagDefaulGatewayIp:
				sprintf(str, "%d.%d.%d.%d",
						data[0], data[1], data[2], data[3]);
				break;
			default:
				memcpy(str, data, length);
				str[length] = '\0';
		}
		if (verbose)
			printf(" %s=%s\n", name, str);
		curval = getenv(name);
		if (!curval || strcmp(curval, str)) {
#ifdef CONFIG_HAS_UID
			forceenv(name, str);
#else
			setenv(name, str);
#endif
			changed = 1;
		}
		data += length;
	}
	return changed;
}

int do_spi_flash_sysinfo(int verbose)
{
  image_header_t *hdr = NULL;
  int changed = 0;
  
  if (!flash) 
    {
      unsigned int bus = 0;
      unsigned int cs = 0;
      flash = spi_flash_probe(0, 0, CONFIG_SF_DEFAULT_SPEED, CONFIG_SF_DEFAULT_MODE);
      if (!flash) 
        printf(" flash probe failed bus=%u cs=%u %uHz mode=%u\n", bus, cs, CONFIG_SF_DEFAULT_SPEED, CONFIG_SF_DEFAULT_MODE);
      else if (verbose)
        printf(" %u KiB %s at %u:%u is now current device\n", flash->size >> 10, flash->name, bus, cs);
    }

  if (flash)
    {
      hdr = load_sysinfo_image(flash, verbose);
      if (hdr)
        changed |= update_sysinfo_from_image(image_get_data(hdr), image_get_data_size(hdr), verbose);
    }

  if (!hdr)
    printf("*** Warning: invalid system information\n");

  // Set variables so that Linux apps know what default IP config is
  changed |= setenv_default("defaultipaddr",    CONFIG_DEFAULT_IPADDR);
  changed |= setenv_default("defaultnetmask",   CONFIG_DEFAULT_NETMASK);
  changed |= setenv_default("defaultgatewayip", CONFIG_DEFAULT_GATEWAYIP);

  // Set default ethaddr, ipaddr, netmask and gatewayip if applicable
  changed |= setenv_default("ethaddr",     CONFIG_DEFAULT_ETHADDR);
  changed |= setenv_default("ipaddr",      getenv("defaultipaddr"));
  changed |= setenv_default("netmask",     getenv("defaultnetmask"));
  changed |= setenv_default("gatewayip",   getenv("defaultgatewayip"));

  if (changed) 
    {
      printf(" saving system information environment variables\n");
      saveenv();
      saveenv();
    }
  
  return 0;
}

static int do_spi_flash_read_write(int argc, char *argv[])
{
	unsigned long addr;
	unsigned long offset;
	unsigned long len;
	void *buf;
	char *endp;
	int ret;

	if (argc < 4)
		goto usage;

	addr = simple_strtoul(argv[1], &endp, 16);
	if (*argv[1] == 0 || *endp != 0)
		goto usage;
	offset = simple_strtoul(argv[2], &endp, 16);
	if (*argv[2] == 0 || *endp != 0)
		goto usage;
	len = simple_strtoul(argv[3], &endp, 16);
	if (*argv[3] == 0 || *endp != 0)
		goto usage;

	buf = map_physmem(addr, len, MAP_WRBACK);
	if (!buf) {
		puts("Failed to map physical memory\n");
		return 1;
	}

	if (strcmp(argv[0], "read") == 0)
		ret = spi_flash_read(flash, offset, len, buf);
	else
		ret = spi_flash_write(flash, offset, len, buf);

	unmap_physmem(buf, len);

	if (ret) {
		printf("SPI flash %s failed\n", argv[0]);
		return 1;
	}

	return 0;

usage:
	printf("Usage: sf %s addr offset len\n", argv[0]);
	return 1;
}

__attribute__((unused))
static int do_spi_flash_erase(int argc, char *argv[])
{
	unsigned long offset;
	unsigned long len;
	char *endp;
	int ret;

	if (argc < 3)
		goto usage;

	offset = simple_strtoul(argv[1], &endp, 16);
	if (*argv[1] == 0 || *endp != 0)
		goto usage;
	len = simple_strtoul(argv[2], &endp, 16);
	if (*argv[2] == 0 || *endp != 0)
		goto usage;

	ret = spi_flash_erase(flash, offset, len);
	if (ret) {
		printf("SPI flash %s failed\n", argv[0]);
		return 1;
	}

	return 0;

usage:
	puts("Usage: sf erase offset len\n");
	return 1;
}

static int do_spi_flash(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	const char *cmd;

	/* need at least two arguments */
	if (argc < 2)
		goto usage;

	cmd = argv[1];

	if (strcmp(cmd, "probe") == 0)
		return do_spi_flash_probe(argc - 1, argv + 1);

	if ((strcmp(cmd, "get") == 0) && (argc == 3) &&
        (strcmp(argv[2], "sysinfo") == 0)) {
		return do_spi_flash_sysinfo(1);
	}

	/* The remaining commands require a selected device */
	if (!flash) {
		puts("No SPI flash selected. Please run `sf probe'\n");
		return 1;
	}

#ifdef CONFIG_SF_READONLY
    if (strcmp(cmd, "read") == 0)
        return do_spi_flash_read_write(argc - 1, argv + 1);
#else
	if (strcmp(cmd, "read") == 0 || strcmp(cmd, "write") == 0)
		return do_spi_flash_read_write(argc - 1, argv + 1);
	if (strcmp(cmd, "erase") == 0)
		return do_spi_flash_erase(argc - 1, argv + 1);
#endif

usage:
	cmd_usage(cmdtp);
	return 1;
}

U_BOOT_CMD(
	sf,	5,	1,	do_spi_flash,
	"SPI flash sub-system",
	"probe [bus:]cs [hz] [mode]	- init flash device on given SPI bus\n"
	"				  and chip select\n"
	"sf read addr offset len 	- read `len' bytes starting at\n"
	"				  `offset' to memory at `addr'\n"
#ifndef CONFIG_SF_READONLY
	"sf write addr offset len	- write `len' bytes from memory\n"
	"				  at `addr' to flash at `offset'\n"
	"sf erase offset len		- erase `len' bytes from `offset'\n"
#endif
	"sf get sysinfo			- get System Information"
);
